package com.sun.xml.internal.ws.policy.sourcemodel;

import java.util.ArrayList;
import com.sun.xml.internal.ws.policy.spi.AssertionCreationException;
import com.sun.xml.internal.ws.policy.PolicyAssertion;
import java.util.List;
import java.util.Queue;
import com.sun.xml.internal.ws.policy.AssertionSet;
import com.sun.xml.internal.ws.policy.Policy;
import java.util.Iterator;
import java.util.Collections;
import com.sun.xml.internal.ws.policy.privateutil.LocalizationMessages;
import java.util.HashMap;
import com.sun.xml.internal.ws.policy.privateutil.PolicyUtils;
import java.util.LinkedList;
import java.util.Collection;
import com.sun.xml.internal.ws.policy.PolicyException;
import java.util.Map;
import com.sun.xml.internal.ws.policy.spi.PolicyAssertionCreator;
import com.sun.xml.internal.ws.policy.privateutil.PolicyLogger;

public class PolicyModelTranslator
{
    private static final PolicyLogger LOGGER;
    private static final PolicyAssertionCreator defaultCreator;
    private final Map<String, PolicyAssertionCreator> assertionCreators;
    
    private PolicyModelTranslator() throws PolicyException {
        this(null);
    }
    
    protected PolicyModelTranslator(Collection<PolicyAssertionCreator> creators) throws PolicyException {
        PolicyModelTranslator.LOGGER.entering(new Object[] { creators });
        Collection<PolicyAssertionCreator> allCreators = new LinkedList();
        PolicyAssertionCreator[] discoveredCreators;
        for (PolicyAssertionCreator creator : discoveredCreators = (PolicyAssertionCreator[])PolicyUtils.ServiceProvider.load(PolicyAssertionCreator.class)) {
            allCreators.add(creator);
        }
        if (creators != null) {
            for (PolicyAssertionCreator creator2 : creators) {
                allCreators.add(creator2);
            }
        }
        Map<String, PolicyAssertionCreator> pacMap = new HashMap();
        for (PolicyAssertionCreator creator3 : allCreators) {
            String[] supportedURIs = creator3.getSupportedDomainNamespaceURIs();
            String creatorClassName = creator3.getClass().getName();
            if (supportedURIs == null || supportedURIs.length == 0) {
                PolicyModelTranslator.LOGGER.warning(LocalizationMessages.WSP_0077_ASSERTION_CREATOR_DOES_NOT_SUPPORT_ANY_URI(creatorClassName));
            }
            else {
                int length2 = supportedURIs.length;
                int j = 0;
                while (j < length2) {
                    String supportedURI = supportedURIs[j];
                    PolicyModelTranslator.LOGGER.config(LocalizationMessages.WSP_0078_ASSERTION_CREATOR_DISCOVERED(creatorClassName, supportedURI));
                    if (supportedURI == null || supportedURI.length() == 0) {
                        throw (PolicyException)PolicyModelTranslator.LOGGER.logSevereException(new PolicyException(LocalizationMessages.WSP_0070_ERROR_REGISTERING_ASSERTION_CREATOR(creatorClassName)));
                    }
                    PolicyAssertionCreator oldCreator = (PolicyAssertionCreator)pacMap.put(supportedURI, creator3);
                    if (oldCreator != null) {
                        throw (PolicyException)PolicyModelTranslator.LOGGER.logSevereException(new PolicyException(LocalizationMessages.WSP_0071_ERROR_MULTIPLE_ASSERTION_CREATORS_FOR_NAMESPACE(supportedURI, oldCreator.getClass().getName(), creator3.getClass().getName())));
                    }
                    ++j;
                }
            }
        }
        this.assertionCreators = Collections.unmodifiableMap(pacMap);
        PolicyModelTranslator.LOGGER.exiting();
    }
    
    public static PolicyModelTranslator getTranslator() throws PolicyException {
        return new PolicyModelTranslator();
    }
    
    public Policy translate(PolicySourceModel model) throws PolicyException {
        PolicyModelTranslator.LOGGER.entering(new Object[] { model });
        if (model == null) {
            throw (PolicyException)PolicyModelTranslator.LOGGER.logSevereException(new PolicyException(LocalizationMessages.WSP_0043_POLICY_MODEL_TRANSLATION_ERROR_INPUT_PARAM_NULL()));
        }
        PolicySourceModel localPolicyModelCopy;
        try {
            localPolicyModelCopy = model.clone();
        }
        catch (CloneNotSupportedException e) {
            throw (PolicyException)PolicyModelTranslator.LOGGER.logSevereException(new PolicyException(LocalizationMessages.WSP_0016_UNABLE_TO_CLONE_POLICY_SOURCE_MODEL(), e));
        }
        String policyId = localPolicyModelCopy.getPolicyId();
        String policyName = localPolicyModelCopy.getPolicyName();
        Collection<AssertionSet> alternatives = this.createPolicyAlternatives(localPolicyModelCopy);
        PolicyModelTranslator.LOGGER.finest(LocalizationMessages.WSP_0052_NUMBER_OF_ALTERNATIVE_COMBINATIONS_CREATED(Integer.valueOf(alternatives.size())));
        Policy policy = null;
        if (alternatives.size() == 0) {
            policy = Policy.createNullPolicy(model.getNamespaceVersion(), policyName, policyId);
            PolicyModelTranslator.LOGGER.finest(LocalizationMessages.WSP_0055_NO_ALTERNATIVE_COMBINATIONS_CREATED());
        }
        else if (alternatives.size() == 1 && ((AssertionSet)alternatives.iterator().next()).isEmpty()) {
            policy = Policy.createEmptyPolicy(model.getNamespaceVersion(), policyName, policyId);
            PolicyModelTranslator.LOGGER.finest(LocalizationMessages.WSP_0026_SINGLE_EMPTY_ALTERNATIVE_COMBINATION_CREATED());
        }
        else {
            policy = Policy.createPolicy(model.getNamespaceVersion(), policyName, policyId, alternatives);
            PolicyModelTranslator.LOGGER.finest(LocalizationMessages.WSP_0057_N_ALTERNATIVE_COMBINATIONS_M_POLICY_ALTERNATIVES_CREATED(Integer.valueOf(alternatives.size()), Integer.valueOf(policy.getNumberOfAssertionSets())));
        }
        PolicyModelTranslator.LOGGER.exiting(policy);
        return policy;
    }
    
    private Collection<AssertionSet> createPolicyAlternatives(PolicySourceModel model) throws PolicyException {
        PolicyModelTranslator.ContentDecomposition decomposition = new PolicyModelTranslator.ContentDecomposition();
        Queue<PolicyModelTranslator.RawPolicy> policyQueue = new LinkedList();
        Queue<Collection<ModelNode>> contentQueue = new LinkedList();
        PolicyModelTranslator.RawPolicy processedPolicy;
        PolicyModelTranslator.RawPolicy rootPolicy = processedPolicy = new PolicyModelTranslator.RawPolicy(model.getRootNode(), new LinkedList());
        do {
            Collection<ModelNode> processedContent = processedPolicy.originalContent;
            do {
                this.decompose(processedContent, decomposition);
                if (decomposition.exactlyOneContents.isEmpty()) {
                    PolicyModelTranslator.RawAlternative alternative = new PolicyModelTranslator.RawAlternative(decomposition.assertions);
                    processedPolicy.alternatives.add(alternative);
                    if (!alternative.allNestedPolicies.isEmpty()) {
                        policyQueue.addAll(alternative.allNestedPolicies);
                    }
                    else {
                        continue;
                    }
                }
                else {
                    Collection<Collection<ModelNode>> combinations = PolicyUtils.Collections.combine(decomposition.assertions, decomposition.exactlyOneContents, false);
                    if (combinations != null && !combinations.isEmpty()) {
                        contentQueue.addAll(combinations);
                    }
                    else {
                        continue;
                    }
                }
            } while ((processedContent = (Collection)contentQueue.poll()) != null);
        } while ((processedPolicy = (PolicyModelTranslator.RawPolicy)policyQueue.poll()) != null);
        Collection<AssertionSet> assertionSets = new LinkedList();
        for (PolicyModelTranslator.RawAlternative rootAlternative : rootPolicy.alternatives) {
            Collection<AssertionSet> normalizedAlternatives = this.normalizeRawAlternative(rootAlternative);
            assertionSets.addAll(normalizedAlternatives);
        }
        return assertionSets;
    }
    
    private void decompose(Collection<ModelNode> content, PolicyModelTranslator.ContentDecomposition decomposition) throws PolicyException {
        decomposition.reset();
        Queue<ModelNode> allContentQueue = new LinkedList(content);
        ModelNode node;
        while ((node = (ModelNode)allContentQueue.poll()) != null) {
            switch (node.getType()) {
                case POLICY:
                case ALL: {
                    allContentQueue.addAll(node.getChildren());
                    continue;
                }
                case POLICY_REFERENCE: {
                    allContentQueue.addAll(getReferencedModelRootNode(node).getChildren());
                    continue;
                }
                case EXACTLY_ONE: {
                    decomposition.exactlyOneContents.add(this.expandsExactlyOneContent(node.getChildren()));
                    continue;
                }
                case ASSERTION: {
                    decomposition.assertions.add(node);
                    continue;
                }
                default: {
                    throw (PolicyException)PolicyModelTranslator.LOGGER.logSevereException(new PolicyException(LocalizationMessages.WSP_0007_UNEXPECTED_MODEL_NODE_TYPE_FOUND(node.getType())));
                }
            }
        }
    }
    
    private static ModelNode getReferencedModelRootNode(ModelNode policyReferenceNode) throws PolicyException {
        PolicySourceModel referencedModel = policyReferenceNode.getReferencedModel();
        if (referencedModel == null) {
            PolicyReferenceData refData = policyReferenceNode.getPolicyReferenceData();
            if (refData == null) {
                throw (PolicyException)PolicyModelTranslator.LOGGER.logSevereException(new PolicyException(LocalizationMessages.WSP_0041_POLICY_REFERENCE_NODE_FOUND_WITH_NO_POLICY_REFERENCE_IN_IT()));
            }
            throw (PolicyException)PolicyModelTranslator.LOGGER.logSevereException(new PolicyException(LocalizationMessages.WSP_0010_UNEXPANDED_POLICY_REFERENCE_NODE_FOUND_REFERENCING(refData.getReferencedModelUri())));
        }
        else {
            return referencedModel.getRootNode();
        }
    }
    
    private Collection<ModelNode> expandsExactlyOneContent(Collection<ModelNode> content) throws PolicyException {
        Collection<ModelNode> result = new LinkedList();
        Queue<ModelNode> eoContentQueue = new LinkedList(content);
        ModelNode node;
        while ((node = (ModelNode)eoContentQueue.poll()) != null) {
            switch (node.getType()) {
                case POLICY:
                case ALL:
                case ASSERTION: {
                    result.add(node);
                    continue;
                }
                case POLICY_REFERENCE: {
                    result.add(getReferencedModelRootNode(node));
                    continue;
                }
                case EXACTLY_ONE: {
                    eoContentQueue.addAll(node.getChildren());
                    continue;
                }
                default: {
                    throw (PolicyException)PolicyModelTranslator.LOGGER.logSevereException(new PolicyException(LocalizationMessages.WSP_0001_UNSUPPORTED_MODEL_NODE_TYPE(node.getType())));
                }
            }
        }
        return result;
    }
    
    private List<AssertionSet> normalizeRawAlternative(PolicyModelTranslator.RawAlternative alternative) throws AssertionCreationException, PolicyException {
        List<PolicyAssertion> normalizedContentBase = new LinkedList();
        Collection<List<PolicyAssertion>> normalizedContentOptions = new LinkedList();
        if (!alternative.nestedAssertions.isEmpty()) {
            Queue<PolicyModelTranslator.RawAssertion> nestedAssertionsQueue = new LinkedList(alternative.nestedAssertions);
            PolicyModelTranslator.RawAssertion rawAssertion;
            while ((rawAssertion = (PolicyModelTranslator.RawAssertion)nestedAssertionsQueue.poll()) != null) {
                List<PolicyAssertion> normalized = this.normalizeRawAssertion(rawAssertion);
                if (normalized.size() == 1) {
                    normalizedContentBase.addAll(normalized);
                }
                else {
                    normalizedContentOptions.add(normalized);
                }
            }
        }
        List<AssertionSet> options = new LinkedList();
        if (normalizedContentOptions.isEmpty()) {
            options.add(AssertionSet.createAssertionSet(normalizedContentBase));
            return options;
        }
        Collection<Collection<PolicyAssertion>> contentCombinations = PolicyUtils.Collections.combine(normalizedContentBase, normalizedContentOptions, true);
        for (Collection<PolicyAssertion> contentOption : contentCombinations) {
            options.add(AssertionSet.createAssertionSet(contentOption));
        }
        return options;
    }
    
    private List<PolicyAssertion> normalizeRawAssertion(PolicyModelTranslator.RawAssertion assertion) throws AssertionCreationException, PolicyException {
        List<PolicyAssertion> parameters;
        if (assertion.parameters.isEmpty()) {
            parameters = null;
        }
        else {
            parameters = new ArrayList(assertion.parameters.size());
            for (ModelNode parameterNode : assertion.parameters) {
                parameters.add(this.createPolicyAssertionParameter(parameterNode));
            }
        }
        List<AssertionSet> nestedAlternatives = new LinkedList();
        if (assertion.nestedAlternatives != null && !assertion.nestedAlternatives.isEmpty()) {
            Queue<PolicyModelTranslator.RawAlternative> nestedAlternativeQueue = new LinkedList(assertion.nestedAlternatives);
            PolicyModelTranslator.RawAlternative rawAlternative;
            while ((rawAlternative = (PolicyModelTranslator.RawAlternative)nestedAlternativeQueue.poll()) != null) {
                nestedAlternatives.addAll(this.normalizeRawAlternative(rawAlternative));
            }
        }
        List<PolicyAssertion> assertionOptions = new LinkedList();
        boolean nestedAlternativesAvailable = !nestedAlternatives.isEmpty();
        if (nestedAlternativesAvailable) {
            for (AssertionSet nestedAlternative : nestedAlternatives) {
                assertionOptions.add(this.createPolicyAssertion(assertion.originalNode.getNodeData(), parameters, nestedAlternative));
            }
            return assertionOptions;
        }
        assertionOptions.add(this.createPolicyAssertion(assertion.originalNode.getNodeData(), parameters, null));
        return assertionOptions;
    }
    
    private PolicyAssertion createPolicyAssertionParameter(ModelNode parameterNode) throws AssertionCreationException, PolicyException {
        if (parameterNode.getType() != ModelNode.Type.ASSERTION_PARAMETER_NODE) {
            throw (PolicyException)PolicyModelTranslator.LOGGER.logSevereException(new PolicyException(LocalizationMessages.WSP_0065_INCONSISTENCY_IN_POLICY_SOURCE_MODEL(parameterNode.getType())));
        }
        List<PolicyAssertion> childParameters = null;
        if (parameterNode.hasChildren()) {
            childParameters = new ArrayList(parameterNode.childrenSize());
            for (ModelNode childParameterNode : parameterNode) {
                childParameters.add(this.createPolicyAssertionParameter(childParameterNode));
            }
        }
        return this.createPolicyAssertion(parameterNode.getNodeData(), childParameters, null);
    }
    
    private PolicyAssertion createPolicyAssertion(AssertionData data, Collection<PolicyAssertion> assertionParameters, AssertionSet nestedAlternative) throws AssertionCreationException {
        String assertionNamespace = data.getName().getNamespaceURI();
        PolicyAssertionCreator domainSpecificPAC = (PolicyAssertionCreator)this.assertionCreators.get(assertionNamespace);
        if (domainSpecificPAC == null) {
            return PolicyModelTranslator.defaultCreator.createAssertion(data, assertionParameters, nestedAlternative, null);
        }
        return domainSpecificPAC.createAssertion(data, assertionParameters, nestedAlternative, PolicyModelTranslator.defaultCreator);
    }
    
    static ModelNode access$000(ModelNode x0) throws PolicyException {
        return getReferencedModelRootNode(x0);
    }
    
    static {
        LOGGER = PolicyLogger.getLogger(PolicyModelTranslator.class);
        defaultCreator = new DefaultPolicyAssertionCreator();
    }
    
    private static final class ContentDecomposition
    {
        final List<Collection<ModelNode>> exactlyOneContents;
        final List<ModelNode> assertions;
        
        private ContentDecomposition() {
            this.exactlyOneContents = new LinkedList();
            this.assertions = new LinkedList();
        }
        
        void reset() {
            this.exactlyOneContents.clear();
            this.assertions.clear();
        }
    }
    
    private static final class RawAssertion
    {
        ModelNode originalNode;
        Collection<PolicyModelTranslator.RawAlternative> nestedAlternatives;
        final Collection<ModelNode> parameters;
        
        RawAssertion(ModelNode originalNode, Collection<ModelNode> parameters) {
            this.nestedAlternatives = null;
            this.parameters = parameters;
            this.originalNode = originalNode;
        }
    }
    
    private static final class RawAlternative
    {
        private static final PolicyLogger LOGGER;
        final List<PolicyModelTranslator.RawPolicy> allNestedPolicies;
        final Collection<PolicyModelTranslator.RawAssertion> nestedAssertions;
        
        RawAlternative(Collection<ModelNode> assertionNodes) throws PolicyException {
            this.allNestedPolicies = new LinkedList();
            this.nestedAssertions = new LinkedList();
            for (ModelNode node : assertionNodes) {
                PolicyModelTranslator.RawAssertion assertion = new PolicyModelTranslator.RawAssertion(node, new LinkedList());
                this.nestedAssertions.add(assertion);
                for (ModelNode assertionNodeChild : assertion.originalNode.getChildren()) {
                    switch (assertionNodeChild.getType()) {
                        case ASSERTION_PARAMETER_NODE: {
                            assertion.parameters.add(assertionNodeChild);
                            continue;
                        }
                        case POLICY:
                        case POLICY_REFERENCE: {
                            if (assertion.nestedAlternatives == null) {
                                assertion.nestedAlternatives = new LinkedList();
                                PolicyModelTranslator.RawPolicy nestedPolicy;
                                if (assertionNodeChild.getType() == ModelNode.Type.POLICY) {
                                    nestedPolicy = new PolicyModelTranslator.RawPolicy(assertionNodeChild, assertion.nestedAlternatives);
                                }
                                else {
                                    nestedPolicy = new PolicyModelTranslator.RawPolicy(PolicyModelTranslator.access$000(assertionNodeChild), assertion.nestedAlternatives);
                                }
                                this.allNestedPolicies.add(nestedPolicy);
                                continue;
                            }
                            else {
                                throw (PolicyException)PolicyModelTranslator.RawAlternative.LOGGER.logSevereException(new PolicyException(LocalizationMessages.WSP_0006_UNEXPECTED_MULTIPLE_POLICY_NODES()));
                            }
                        }
                        default: {
                            throw (PolicyException)PolicyModelTranslator.RawAlternative.LOGGER.logSevereException(new PolicyException(LocalizationMessages.WSP_0008_UNEXPECTED_CHILD_MODEL_TYPE(assertionNodeChild.getType())));
                        }
                    }
                }
            }
        }
        
        static {
            LOGGER = PolicyLogger.getLogger(PolicyModelTranslator.RawAlternative.class);
        }
    }
    
    private static final class RawPolicy
    {
        final Collection<ModelNode> originalContent;
        final Collection<PolicyModelTranslator.RawAlternative> alternatives;
        
        RawPolicy(ModelNode policyNode, Collection<PolicyModelTranslator.RawAlternative> alternatives) {
            this.originalContent = policyNode.getChildren();
            this.alternatives = alternatives;
        }
    }
}
